/*

LibHFZ, a library to load and save HF2/HFZ compressed heightfield files.

Version: 1.1.1
Released: 9th of April 2007

Copyright (C) 2007 Aaron Torpy (aaron@bundysoft.com)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

For more information about LibHFZ, please consult:
http://www.bundysoft.com/docs/doku.php?id=libhfz

*/

#include <stdio.h>
#include <memory.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
	

#include "libhfz.h"
#include "zlib\include\zlib.h"
#include "zlib\include\zconf.h"


//
//
// High-level HFZ functions
//
//


long hfzSave(const char* lpFileName, long Format, hfzHeader& fh, float* pData) {

	// check format
	switch(Format) {
	default:
		return LIBHFZ_ERROR_INVALID_FORMAT;
	case LIBHFZ_FORMAT_HF2:
	case LIBHFZ_FORMAT_HF2_GZ:
		break;
	}

	// check settings
	if(fh.Precis<=0) {
		return LIBHFZ_ERROR_INVALID_PRECIS;
	}
	if(fh.TileSize<8 || fh.TileSize>65535) {
		return LIBHFZ_ERROR_INVALID_TILE_SIZE;
	}
	if(fh.nx<=0 || fh.ny<=0) {
		return LIBHFZ_ERROR_INVALID_MAP_SIZE;
	}
	
	// determine number of blocks
	long nBlocksX = fh.nx / fh.TileSize;
	long nBlocksY = fh.ny / fh.TileSize;
	if(fh.nx%fh.TileSize)	nBlocksX++;
	if(fh.ny%fh.TileSize)	nBlocksY++;
	long BlockCount = nBlocksX * nBlocksY;

	// open file for writing
	hfzFile* fs;
	if(!(fs = hfzOpen(lpFileName, Format, "wb"))) {
		return LIBHFZ_ERROR_OPEN_FAILED;
	}

	// write header
	if(LIBHFZ_STATUS_OK!=hfzWriteHeader(fs, fh)) {
		hfzClose(fs);
		return LIBHFZ_ERROR_WRITE_FAILED;
	}

	// writing tiles
	long rval = LIBHFZ_STATUS_OK;
	for(long bj = 0; bj<nBlocksY && rval==LIBHFZ_STATUS_OK; bj++) {
	for(long bi = 0; bi<nBlocksX && rval==LIBHFZ_STATUS_OK; bi++) {
		rval = hfzWriteTile(fs, fh, bi, bj, pData);
	}
	}

	// close file
	hfzClose(fs);

	return rval;
}

long hfzLoad(const char* lpFileName, hfzHeader& fh, float *pData) {

	if(!pData)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	hfzHeader_Reset(fh);

	long rval;

	hfzFile* fs;
	fs = hfzOpen(lpFileName, LIBHFZ_FORMAT_HF2_GZ, "rb"); // always treat as gzip; zlib handles raw too
	if(!fs) {
		return LIBHFZ_ERROR_OPEN_FAILED;
	}

	rval = hfzReadHeader(fs, fh);
	if(LIBHFZ_STATUS_OK!=rval) {
		hfzClose(fs);
		return rval;
	}

	// shorthand (shorter hand, anyhow)
	long nx = fh.nx;
	long ny = fh.ny;
	long TileSize = fh.TileSize;

	// debugging
	//TempStr.Format("VersionNo = %d\r\nnx = %d\r\nny = %d\r\nBlockSize = %d\r\nPrecis = %.2f\r\nHorizScale = %.2f\r\nExtHeaderLength = %d", FH_HFZ.FileVersionNo, FH_HFZ.nx, FH_HFZ.ny, FH_HFZ.BlockSize, FH_HFZ.Precis, FH_HFZ.HorizScale, FH_HFZ.ExtHeaderLength);
	//AfxMessageBox(TempStr);

	// determine number of blocks
	long nBlocksX = nx / TileSize;
	long nBlocksY = ny / TileSize;
	if(nx%TileSize)	nBlocksX++;
	if(ny%TileSize)	nBlocksY++;
	long BlockCount = nBlocksX * nBlocksY;

	// start reading tiles
	bool ErrorFlag = false;

	rval = LIBHFZ_STATUS_OK;
	for(long bj = 0; bj<nBlocksY && rval == LIBHFZ_STATUS_OK; bj++) {
		for(long bi = 0; bi<nBlocksX && rval == LIBHFZ_STATUS_OK; bi++) {

			rval = hfzReadTile(fs, fh, bi, bj, pData);
		}
	}

	hfzClose(fs);

	return rval;
}

long hfzLoadEx(const char* lpFileName, hfzHeader& fh, float **h_pData) {

	hfzHeader_Reset(fh);
	long rval = hfzReadHeader2(lpFileName, fh);
	if(LIBHFZ_STATUS_OK!=rval) {
        return rval;		
	}

	// allocate buffer
	float* pData = (float*)calloc(fh.nx * fh.ny, 4);
	if(!pData) {
		hfzHeader_Reset(fh);
		return LIBHFZ_ERROR_ALLOC_FAILED;
	}

	// load 
	rval = hfzLoad(lpFileName, fh, pData);
	if(LIBHFZ_STATUS_OK!=rval) {
		hfzHeader_Reset(fh);
		free(pData);
		return rval;
	}

	*h_pData = pData;
	return LIBHFZ_STATUS_OK;
}

//
//
// Medium-level HFZ functions
//
//

long hfzWriteHeader(hfzFile* fs, hfzHeader& fh) {

	// copy header into buffer
	char HeaderBuf[28];
	long HeaderBufPos = 0;
	sprintf(HeaderBuf+HeaderBufPos, "HF2"); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.FileVersionNo), 2); HeaderBufPos+=2;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.nx), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.ny), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.TileSize), 2); HeaderBufPos+=2;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.Precis), 4); HeaderBufPos+=4;
	memcpy(HeaderBuf+HeaderBufPos, &(fh.HorizScale), 4); HeaderBufPos+=4;

	// put extended header into a buffer
	char* pExtHeaderData = 0;
	long rval = hfzHeader_EncodeExtHeaderBuf(fh, &pExtHeaderData);
	if(rval<0) return rval;

	// now write in header length of ext header
	memcpy(HeaderBuf+HeaderBufPos, &(fh.ExtHeaderLength), 4); HeaderBufPos+=4;

	// write header
	if(28!=hfzWrite(fs, HeaderBuf, 28)) 
		return LIBHFZ_ERROR_WRITE_HEADER_FAILED;

	// write extended header
	if(pExtHeaderData) {
		if(fh.ExtHeaderLength!=hfzWrite(fs, pExtHeaderData, fh.ExtHeaderLength)) {
			free(pExtHeaderData);
			return LIBHFZ_ERROR_WRITE_EXTHEAD_FAILED;
		}
		free(pExtHeaderData);
	}

	return LIBHFZ_STATUS_OK;
}

long hfzReadHeader(hfzFile* fs, hfzHeader& fh) {

	// initialising empty
	hfzHeader_Init(fh, 0, 0, 0, 0, 0, 0);

	// read the core header
	char HeaderBuf[28];
	if(28!=hfzRead(fs, HeaderBuf, 28)) {
		return LIBHFZ_ERROR_READ_FAILED;
	}

	// check for "HF2\0" in leading 4 bytes
	HeaderBuf[3] = 0;
	if(strcmp(HeaderBuf, "HF2")) {
		return LIBHFZ_ERROR_INVALID_FORMAT;
	}

	long HeaderPos = 4;
	memcpy(&(fh.FileVersionNo), HeaderBuf+HeaderPos, 2); HeaderPos+=2;
	memcpy(&(fh.nx), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.ny), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.TileSize), HeaderBuf+HeaderPos, 2); HeaderPos+=2;
	memcpy(&(fh.Precis), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.HorizScale), HeaderBuf+HeaderPos, 4); HeaderPos+=4;
	memcpy(&(fh.ExtHeaderLength), HeaderBuf+HeaderPos, 4); HeaderPos+=4;

	unsigned long hl = fh.ExtHeaderLength;

    if(hl) {

		char* pBuf;
		if(!(pBuf = (char*)calloc(hl, 1))) {
			return LIBHFZ_ERROR_ALLOC_FAILED;
		}

        if(hl!=hfzRead(fs, pBuf, hl)) {
			free(pBuf);
			return LIBHFZ_ERROR_READ_FAILED;
		}

		int rval = hfzHeader_DecodeExtHeaderBuf(fh, pBuf);
		if(rval<0) {
			free(pBuf);
			return rval;
		}
	}

	return LIBHFZ_STATUS_OK;
}

long hfzReadHeader2(const char* lpFileName, hfzHeader& fh) {
	hfzFile* fs;
	if(!(fs=hfzOpen(lpFileName, LIBHFZ_FORMAT_HF2_GZ, "rb"))) {
		return LIBHFZ_ERROR_OPEN_FAILED;
	}

	if(LIBHFZ_STATUS_OK!=hfzReadHeader(fs, fh)) {
		return LIBHFZ_ERROR_READ_FAILED;
	}

	hfzClose(fs);

	return LIBHFZ_STATUS_OK;
}


long hfzReadTile(hfzFile* fs, hfzHeader& fh, long TileX, long TileY, float* pMapData) {

	long i1, i2, j1, j2, i, j;

	char c;
	short s;
	long li;

	long TileSize = fh.TileSize;
	long nx = fh.nx;
	long ny = fh.ny;

	i1 = TileX * TileSize;
	j1 = TileY * TileSize;
	i2 = i1 + TileSize;
	j2 = j1 + TileSize;

	if(i2>=nx)	i2 = nx;
	if(j2>=ny)	j2 = ny;

	long bnx = i2 - i1;
	long bny = j2 - j1;

	// read vert offset and sale
	float VertOffset;
	float VertScale;
	if(4!=hfzRead(fs, &VertScale, 4))
		return LIBHFZ_ERROR_READ_FAILED;
	if(4!=hfzRead(fs, &VertOffset, 4))
		return LIBHFZ_ERROR_READ_FAILED;

	// debugging
	//TempStr.Format("VertScale = %f\r\nVertOffset = %f", VertOffset, VertScale);
	//AfxMessageBox(TempStr);

	for(j=j1; j<j2; j++) {

		char LineDepth; // 1, 2, or 4
		long FirstVal;

		if(1!=hfzRead(fs, &LineDepth, 1))
			return LIBHFZ_ERROR_READ_FAILED;
		if(4!=hfzRead(fs, &FirstVal, 4))
			return LIBHFZ_ERROR_READ_FAILED;

		float f = (float)FirstVal * VertScale + VertOffset;

		// set first pixel
		pMapData[i1 + nx * j] = f;

		long LastVal = FirstVal;

		for(i=i1+1; i<i2; i++) {
			switch(LineDepth) {
			case 1:
				if(1!=hfzRead(fs, &c, LineDepth))	return LIBHFZ_ERROR_READ_FAILED;
				li = (long)c;
				break;
			case 2:
				if(2!=hfzRead(fs, &s, LineDepth))	return LIBHFZ_ERROR_READ_FAILED;
				li = (long)s;
				break;
			case 4:
				if(4!=hfzRead(fs, &li, LineDepth))	return LIBHFZ_ERROR_READ_FAILED;
				break;
			}

			f = (float)(li +  LastVal) * VertScale + VertOffset;
			LastVal = li + LastVal;

			pMapData[i + nx * j] = f;
		}
	}
	
	return LIBHFZ_STATUS_OK;
}

long hfzWriteTile(hfzFile* fs, hfzHeader& fh, long TileX, long TileY, float* pMapData) {

	if(!fs)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	if(!pMapData)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	long i, j;
	long TempInt;
	float f;
	float HFmin, HFmax;
	char c;
	short s;
	long i1, j1, i2, j2, bnx, bny;
	long FirstVal;
	float VertScale;
	float VertOffset;

	long nx = fh.nx;
	long ny = fh.ny;
	long TileSize = fh.TileSize;
	float Precis = fh.Precis;


	i1 = TileX * TileSize;
	j1 = TileY * TileSize;
	i2 = i1 + TileSize;
	j2 = j1 + TileSize;

	if(i2>=nx)	i2 = nx;
	if(j2>=ny)	j2 = ny;

	// debugging
	//CString TempStr;
	//TempStr.Format("x = (%d...%d), y = (%d...%d)", i1, i2, j1, j2);
	//AfxMessageBox(TempStr);

	bnx = i2 - i1;
	bny = j2 - j1;

	// prepare buffer for line
	char * pDiffBuf = (char*)calloc(TileSize * 4, 1);
	if(!pDiffBuf) {
		return LIBHFZ_ERROR_ALLOC_FAILED;
	}
	char * pOutBuf = (char*)calloc(TileSize * 4, 1);
	if(!pOutBuf) {
		return LIBHFZ_ERROR_ALLOC_FAILED;
	}

	// get min/max alt in block (used for vert scale)
	HFmin = 0;
	HFmax = 0;
	for(j=j1; j<j2; j++) {
		for(i=i1; i<i2; i++) {
			// find max diff in line

			f = pMapData[i + j * nx];
		
			if(j==j1 && i==i1) {
				HFmin = HFmax = f;
			} else {
				if(f<HFmin)	HFmin = f;
				if(f>HFmin)	HFmax = f;
			}
		}
	}

	// number of int levels required for this block
	float BlockLevels = (HFmax - HFmin) / Precis + 1;

	// debugging
	//TempStr.Format("Block %d,%d: %.1f", bi, bj, BlockLevels);
	//theAPI.ReportError(TempStr);

	// calc scale
	VertScale = (HFmax - HFmin)/(float)BlockLevels;
	VertOffset = HFmin;
	if(VertScale<=0)	VertScale = 1.0f; // this is for niceness

	// write the block scaling
	char ScaleBuf[8];
	memcpy(ScaleBuf, &VertScale, 4);
	memcpy(ScaleBuf+4, &VertOffset, 4);
	if(8!=hfzWrite(fs, ScaleBuf, 8)) {
		return LIBHFZ_ERROR_WRITE_FAILED;
	}

	// save block line-by-line
	for(j=j1; j<j2; j++) {
	
		// get first val
		f = pMapData[i1 + j * nx];
		FirstVal = (long)((f-VertOffset)/VertScale);
		long LastVal = FirstVal;

		// find max diff in line
		long Diff;
		long MaxDev = 0;
		for(i=i1+1; i<i2; i++) {

			// find max diff in line
			f = pMapData[i + j * nx];
		
			TempInt=(long)((f-VertOffset)/VertScale);
			Diff = TempInt - LastVal;

			long zi = i-i1-1;
			memcpy((char*)pDiffBuf+zi*4, &Diff, 4);

			LastVal = TempInt;

			MaxDev = MaxDev>abs(Diff)?MaxDev:abs(Diff);
		}

		// should we use 8, 16 or 32 bit pixels?
		char LineDepth = 4;
		if(MaxDev<=127) {
			LineDepth = 1;
		} else
		if(MaxDev<=32768) {
			LineDepth = 2;
		}

		// write line header
		char BlockBuf[5];
		memcpy(BlockBuf, &LineDepth, 1); // store line depth
		memcpy(BlockBuf+1, &FirstVal, 4); // store first value (32bit precis)
		if(5!=hfzWrite(fs, BlockBuf, 5)) {
			return LIBHFZ_ERROR_WRITE_FAILED;
		}
		
		// now write block
		char* pOutBuf2 = pOutBuf;
		for(i=i1+1; i<i2; i++) {

			long zi = i-i1-1;
			memcpy(&Diff, (char*)pDiffBuf+zi*4, 4);
			switch(LineDepth) {
			case 1:
				c = (char)Diff;
				memcpy(pOutBuf2, &c, LineDepth);
				break;
			case 2:
				s = (short)Diff;
				memcpy(pOutBuf2, &s, LineDepth);
				break;
			case 4:
				memcpy(pOutBuf2, &Diff, LineDepth);
				break;
			}

			pOutBuf2+=LineDepth;
		}

		long len = LineDepth * (i2-i1-1);
		if(len!=hfzWrite(fs, pOutBuf, len))
			return LIBHFZ_ERROR_WRITE_FAILED;
	}

	free(pDiffBuf);
	free(pOutBuf);

	return LIBHFZ_STATUS_OK;
}

//
//
// Low-level HFZ functions
//
//

hfzFile* hfzOpen(const char* lpFileName, long Format, const char* lpMode) {
	
	hfzFile* fs = new hfzFile;

	switch(Format) {
	case LIBHFZ_FORMAT_HF2: 
		if(!(fs->raw = fopen(lpFileName, lpMode))) {
			delete fs;
			return 0;
		}
		break;
	case LIBHFZ_FORMAT_HF2_GZ: 
		if(!(fs->gz = gzopen(lpFileName, lpMode))) {
			delete fs;
			return 0;
		}
		break;
	default: 
		delete fs;
		return 0;
	}

	fs->Format = Format;

	return fs;
}

void hfzClose(hfzFile* fs) {

	if(!fs)
		return;

	switch(fs->Format) {
	case LIBHFZ_FORMAT_HF2:	fclose(fs->raw); break;
	case LIBHFZ_FORMAT_HF2_GZ: gzclose(fs->gz); break;
	}

	delete fs;
}

long hfzRead(hfzFile* fs, void* pData, long len) {
	if(!fs)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	if(!len)
		return 0;

	if(!pData)
		return LIBHFZ_ERROR_INVALID_HANDLE;

    switch(fs->Format) {
	case LIBHFZ_FORMAT_HF2:    return (long)fread(pData, len, 1, fs->raw);
	case LIBHFZ_FORMAT_HF2_GZ: return (long)gzread(fs->gz, pData, len);
	}

	return LIBHFZ_ERROR_INVALID_FORMAT;
}

long hfzWrite(hfzFile* fs, void* pData, long len) {
	if(!fs)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	if(!len)
		return 0;

	if(!pData)
		return LIBHFZ_ERROR_INVALID_HANDLE;

	switch(fs->Format) {
	case LIBHFZ_FORMAT_HF2:    return fwrite(pData, 1, len, fs->raw);
	case LIBHFZ_FORMAT_HF2_GZ: return gzwrite(fs->gz, pData, len);
	}

	return LIBHFZ_ERROR_INVALID_FORMAT;
}

//
//
// HFZ helper functions
//
//


const char* hfzGetErrorStr(long ErrorCode) {
	switch(ErrorCode) {
	case LIBHFZ_ERROR_OPEN_FAILED:			return "Cannot open file";
	case LIBHFZ_ERROR_WRITE_FAILED:			return "Write failed";
	case LIBHFZ_ERROR_READ_FAILED:			return "Read failed";
	case LIBHFZ_ERROR_ALLOC_FAILED:			return "Cannot allocate buffer";
	case LIBHFZ_ERROR_INVALID_PARAM:		return "Invalid parameter";
	case LIBHFZ_ERROR_INVALID_HANDLE:		return "Invalid handle";
	case LIBHFZ_ERROR_INVALID_FORMAT:		return "Invalid format ID";
	case LIBHFZ_ERROR_INVALID_PRECIS:		return "Invalid precision";
	case LIBHFZ_ERROR_INVALID_MAP_SIZE:		return "Invalid map size";
	case LIBHFZ_ERROR_INVALID_TILE_SIZE:	return "Invalid tile size";
	case LIBHFZ_ERROR_INVALID_EXT_HEADER:	return "Invalid extended header";
	case LIBHFZ_ERROR_WRITE_HEADER_FAILED:	return "Error writing header";
	case LIBHFZ_ERROR_WRITE_EXTHEAD_FAILED: return "Error writign extended header";
	}

	return 0;
}

long hfzHeader_Init(hfzHeader &fh, long nx, long ny, unsigned short TileSize, float Precis, float HorizScale, long nExtHeaderBlocks) {

	fh.FileVersionNo = 0; // this is the very first implementation
	fh.nx = nx;
	fh.ny = ny;
	fh.TileSize = TileSize;
	fh.Precis = Precis;
	fh.HorizScale = HorizScale;
	fh.nExtHeaderBlocks = nExtHeaderBlocks;

	try {

		if(nExtHeaderBlocks) {
			fh.pExtHeaderBlocks = (hfzExtHeaderBlock*)calloc(sizeof(hfzExtHeaderBlock), nExtHeaderBlocks);
			if(!fh.pExtHeaderBlocks) {
				return LIBHFZ_ERROR_ALLOC_FAILED;
			}
			for(int i=0; i<nExtHeaderBlocks; i++) {
				// initialising blocks empty
				hfzExtHeaderBlock_InitEx(fh, i, 0, 0, 0, 0);
			}
		} else {
			fh.pExtHeaderBlocks = 0;
		}

	} catch(...) {

		return LIBHFZ_ERROR_INVALID_PARAM;

	}

	return LIBHFZ_STATUS_OK;
}

void hfzHeader_Reset(hfzHeader& fh) {
	fh.FileVersionNo = 0;
	fh.nx = 0;
	fh.ny = 0;
	fh.HorizScale = 1;
	fh.Precis = 0.01f;
	fh.TileSize = 256;
	fh.ExtHeaderLength = 0;

	for(int i=0; i<fh.nExtHeaderBlocks; i++) {
		hfzExtHeaderBlock_Reset(&(fh.pExtHeaderBlocks[i]));
	}
	fh.nExtHeaderBlocks = 0;
	if(fh.pExtHeaderBlocks) {
		free(fh.pExtHeaderBlocks);
		fh.pExtHeaderBlocks = 0;
	}
}

long hfzHeader_EncodeExtHeaderBuf(hfzHeader &fh, char** ppBuf) {

	char* pBuf;

	fh.ExtHeaderLength = 0;

	if(!fh.nExtHeaderBlocks) {
		return LIBHFZ_STATUS_OK;
	}
	
	//calculate overall length
	long extlen = 0;
	for(int i=0; i<fh.nExtHeaderBlocks; i++) {
		extlen += 24 + fh.pExtHeaderBlocks[i].BlockLength;
	}

	if(!extlen) {
		// can't have zero length
		return LIBHFZ_ERROR_INVALID_EXT_HEADER;
	}

	pBuf = (char*)calloc(extlen, 1);
	if(!pBuf) {
		return LIBHFZ_ERROR_ALLOC_FAILED;
	}

	int offset = 0;

	hfzExtHeaderBlock* pBlock;
	for(int i=0; i<fh.nExtHeaderBlocks; i++) {
		pBlock = &(fh.pExtHeaderBlocks[i]);

		// forcing nulls
		pBlock->BlockType[3] = 0;
		pBlock->BlockName[15] = 0;

		memcpy(pBuf+offset, &(pBlock->BlockType), 4); offset+=4;
		memcpy(pBuf+offset, &(pBlock->BlockName), 16); offset+=16;
		memcpy(pBuf+offset, &(pBlock->BlockLength), 4); offset+=4;

		if(pBlock->BlockLength) {
			memcpy(pBuf+offset, pBlock->pBlockData, pBlock->BlockLength); offset+=pBlock->BlockLength;
		}

		// debugging
		/*CString TempStr;
		TempStr.Format("'%s' '%s' %d", pBlock->BlockType, pBlock->BlockName, pBlock->BlockLength);
		if(pBlock->BlockLength) {
			TempStr.Append(" '");
			TempStr.Append((char*)pBlock->pBlockData);
			TempStr.Append("'");
		}
		AfxMessageBox(TempStr);*/
	}

	fh.ExtHeaderLength  = extlen;
	*ppBuf = pBuf;

	return extlen;
}

long hfzHeader_DecodeExtHeaderBuf(hfzHeader &fh, char* pBuf) {

	if(fh.ExtHeaderLength==0)
		return LIBHFZ_STATUS_OK;

	fh.nExtHeaderBlocks = 0;

	// read once to determine number of blocks
	bool DoneFlag = false;
	int offset = 0;
	unsigned long BlockLen;
	do {
		if(offset==fh.ExtHeaderLength) {
			DoneFlag = true;
		} else
		if(fh.ExtHeaderLength-offset<24) {
			// not enough for a complete header
			return LIBHFZ_ERROR_INVALID_EXT_HEADER;
		} else {
			memcpy(&BlockLen, pBuf+offset+20, 4); // getting block length
			offset+=24+BlockLen;
			fh.nExtHeaderBlocks++;
		}
	} while(!DoneFlag);

	//now allocate correct number of blocks
	fh.pExtHeaderBlocks = (hfzExtHeaderBlock*)calloc(sizeof(hfzExtHeaderBlock), fh.nExtHeaderBlocks);
	if(!fh.pExtHeaderBlocks) {
		return LIBHFZ_ERROR_ALLOC_FAILED;
	}

	offset = 0;
	hfzExtHeaderBlock* pBlock;
	for(int i=0; i<fh.nExtHeaderBlocks; i++) {
		pBlock = &(fh.pExtHeaderBlocks[i]);
		memcpy(&(pBlock->BlockType), pBuf+offset, 4); offset+=4;
		memcpy(&(pBlock->BlockName), pBuf+offset, 16); offset+=16;
		memcpy(&(pBlock->BlockLength), pBuf+offset, 4); offset+=4;

		// forcing nulls
		pBlock->BlockType[3] = 0;
		pBlock->BlockName[15] = 0;

		if(pBlock->BlockLength) {
			pBlock->pBlockData = calloc(pBlock->BlockLength, 1);
			memcpy(pBlock->pBlockData, pBuf+offset, pBlock->BlockLength); 
			offset+=pBlock->BlockLength;
		}
	}

	return fh.nExtHeaderBlocks;
}

long hfzExtHeaderBlock_Init(hfzExtHeaderBlock* pBlock, const char* lpBlockType, const char* lpBlockName, unsigned long BlockDataLength, void* pBlockData) {

	for(int i=0; i<4; i++)	pBlock->BlockType[i] = 0;
	for(int i=0; i<16; i++)	pBlock->BlockName[i] = 0;


	if(lpBlockType)	{
		if(strlen(lpBlockType)>=4)
			return LIBHFZ_ERROR_INVALID_PARAM;

		sprintf(pBlock->BlockType, lpBlockType);
	}

	if(lpBlockName)	{
		if(strlen(lpBlockName)>=16)
			return LIBHFZ_ERROR_INVALID_PARAM;

		sprintf(pBlock->BlockName, lpBlockName);
	}
	
	
	pBlock->BlockLength = BlockDataLength;

	if(!pBlock->BlockLength) {
		pBlock->pBlockData = 0;
	} else {
		if(!(pBlock->pBlockData = calloc(pBlock->BlockLength, 1)))
			return LIBHFZ_ERROR_ALLOC_FAILED;

		memcpy(pBlock->pBlockData, pBlockData, pBlock->BlockLength);
	}

	return LIBHFZ_STATUS_OK;
}

long hfzExtHeaderBlock_InitEx(hfzHeader& fh, long BlockID, const char* lpBlockType, const char* lpBlockName, unsigned long BlockDataLength, void* pBlockData) {

	if(BlockID>=fh.nExtHeaderBlocks)
		return LIBHFZ_ERROR_INVALID_PARAM;

	return hfzExtHeaderBlock_Init(&(fh.pExtHeaderBlocks[BlockID]), lpBlockType, lpBlockName, BlockDataLength, pBlockData);
}

void hfzExtHeaderBlock_Reset(hfzExtHeaderBlock* pBlock) {
	for(int i=0; i<4; i++)	pBlock->BlockType[i] = 0;
	for(int i=0; i<16; i++)	pBlock->BlockName[i] = 0;

	if(pBlock->pBlockData) {
		free(pBlock->pBlockData);
		pBlock->pBlockData = 0;
	}

	pBlock->BlockLength = 0;
}